/*  BotikBalance - User's account balance informer for Botik Network 
 *                 for Windows 9x/ME/NT/2000/XP/2003
 *  Copyright (C) 2005 Alexander V. Melnikoff
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public Licensse as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Sorry for the long lines in the sources, I have wide screen... ;-)
 */

/* In this file I didn't use IConnectionPoints and other staff to implement
 * events. I think it's not a right way, but it suits me fine. This method
 * I've borrowed from Microsoft's XMLHTTPRequest and XMLDOMDocument objects
 * implementation.
 * RTFS below for the more information. See 
 *   VOID CBotikBalanceObj::Fire_OnUpdate(VOID)
 *   HRESULT CBotikBalanceObj::get_LastOperationCode(LONG* plRet_)
 */

#include <stdafx.h>
#include <wininet.h>
#include <botikbalanceobj.h>

#include <charsets/koi2win.cpp>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CBotikBalanceObj::CBotikBalanceObj(void)
{
	cfdprintf("CBotikBalanceObj::CBotikBalanceObj");

	m_pITypeInfo        = NULL;
	m_pOnDisconnectSink = NULL;
	m_pOnConnectSink    = NULL;
	m_pOnUpdateSink     = NULL;
	m_cRef              = 1;

	InterlockedIncrement(&g_lDllRefCnt);

	cfdprintf("!CBotikBalanceObj::CBotikBalanceObj");
};

CBotikBalanceObj::~CBotikBalanceObj(void)
{
	cfdprintf("CBotikBalanceObj::~CBotikBalanceObj");
	
	if(m_pITypeInfo) m_pITypeInfo->Release();
	if(m_pOnUpdateSink) m_pOnUpdateSink->Release();
	if(m_pOnConnectSink) m_pOnConnectSink->Release();
	if(m_pOnDisconnectSink) m_pOnDisconnectSink->Release();

	InterlockedDecrement(&g_lDllRefCnt);

	cfdprintf("!CBotikBalanceObj::~CBotikBalanceObj");
};

//////////////////////////////////////////////////////////////////////
// IUnknown
//////////////////////////////////////////////////////////////////////

HRESULT __stdcall CBotikBalanceObj::QueryInterface(const IID& riid_, void** ppv_)
{
	if(riid_ == IID_IUnknown || riid_ == IID_IBotikBalanceObj || riid_ == IID_IDispatch) 
	{
		*ppv_ = static_cast<IBotikBalanceObj*>(this);
	} 
	else 
	{
		*ppv_ = NULL;
		return E_NOINTERFACE;
	};

	reinterpret_cast<IUnknown*>(*ppv_)->AddRef ();

	return(S_OK);
};

ULONG __stdcall CBotikBalanceObj::AddRef(void) 
{
	return(InterlockedIncrement(&m_cRef));
};

ULONG __stdcall CBotikBalanceObj::Release (void) 
{
	if(InterlockedDecrement(&m_cRef) == 0)
	{
		delete this;
		return 0;
	};
	return(m_cRef);
};

//////////////////////////////////////////////////////////////////////
// IDispatch
//////////////////////////////////////////////////////////////////////

HRESULT __stdcall CBotikBalanceObj::GetTypeInfoCount(UINT* pCountTypeInfo_) 
{
	*pCountTypeInfo_ = 1;

	return(S_OK);
};

HRESULT __stdcall CBotikBalanceObj::GetTypeInfo(UINT iTypeInfo_, LCID, ITypeInfo** ppITypeInfo_) 
{
	*ppITypeInfo_ = NULL;
	if(iTypeInfo_ != 0) 
	{
		return(DISP_E_BADINDEX);
	};

	// Call AddRef () and return pointer.
	m_pITypeInfo->AddRef();
	*ppITypeInfo_ = m_pITypeInfo;

	return(S_OK);
};

/* It doesn't support locale. */
HRESULT __stdcall CBotikBalanceObj::GetIDsOfNames(const IID& riid_, OLECHAR** arrayNames_, 
												  UINT cNames_, LCID, DISPID* arrayDispIDs_) 
{
	if(riid_ != IID_NULL) 
	{
		return(DISP_E_UNKNOWNINTERFACE);
	};

	return(m_pITypeInfo->GetIDsOfNames(arrayNames_, cNames_, arrayDispIDs_));
};

/* It doesn't support locale. */
HRESULT __stdcall CBotikBalanceObj::Invoke(DISPID dispidMember_, const IID& riid_, 
										   LCID, WORD wFlags_, DISPPARAMS* pDispParams_, 
										   VARIANT* pvarRes_, EXCEPINFO* pEI_, UINT* pArgErr_) 
{
	if(riid_ != IID_NULL) 
	{
		return(DISP_E_UNKNOWNINTERFACE);
	};
	::SetErrorInfo(0, NULL);

	return(m_pITypeInfo->Invoke(static_cast<IDispatch*>(this), dispidMember_, 
								wFlags_, pDispParams_, pvarRes_, pEI_, pArgErr_));
};

//////////////////////////////////////////////////////////////////////
// IProvideClassInfo
//////////////////////////////////////////////////////////////////////
/*
HRESULT __stdcall CBotikBalanceObj::GetClassInfo(ITypeInfo **ppTI_)
{
	*ppTI_ = NULL;

	// Call AddRef () and return pointer.
	m_pITypeInfo->AddRef();
	*ppTI_ = m_pITypeInfo;

	return(S_OK);
};
*/
//////////////////////////////////////////////////////////////////////
// IBotikBalanceObj
//////////////////////////////////////////////////////////////////////

HRESULT __stdcall CBotikBalanceObj::put_onUpdate([in] IDispatch* pOnUpdateSink_)
{
	cfdprintf(_T("CBotikBalanceObj::put_onUpdate"));

	if(m_pOnUpdateSink)
        m_pOnUpdateSink->Release();

	m_pOnUpdateSink = pOnUpdateSink_;
    if(m_pOnUpdateSink)
        m_pOnUpdateSink->AddRef();

	cfdprintf(_T("!CBotikBalanceObj::put_onUpdate"));

	return(S_OK);
};

HRESULT __stdcall CBotikBalanceObj::put_onConnect([in] IDispatch* pOnConnectSink_)
{
	cfdprintf(_T("CBotikBalanceObj::put_onConnect"));

	if(m_pOnConnectSink)
        m_pOnConnectSink->Release();

	m_pOnConnectSink = pOnConnectSink_;
    if(m_pOnConnectSink)
        m_pOnConnectSink->AddRef();

	cfdprintf(_T("!CBotikBalanceObj::put_onConnect"));

	return(S_OK);
};

HRESULT __stdcall CBotikBalanceObj::put_onDisconnect([in] IDispatch* pOnDisconnectSink_)
{
	cfdprintf(_T("CBotikBalanceObj::put_onDisconnect"));

	if(m_pOnDisconnectSink)
        m_pOnDisconnectSink->Release();

	m_pOnDisconnectSink = pOnDisconnectSink_;
    if(m_pOnDisconnectSink)
        m_pOnDisconnectSink->AddRef();

	cfdprintf(_T("!CBotikBalanceObj::put_onDisconnect"));

	return(S_OK);
};

// XXX 
// Temporary hack. Solve it later...
// See a set of #undefs at the end of following method.
#define __finally	cleanup:
#define __try
#define __leave		goto cleanup
HRESULT __stdcall CBotikBalanceObj::RequestPage(LONG lMethod_, BSTR bsHost_, LONG lPort_, 
												BSTR bsPath_, BSTR bsBody_, BSTR bsHeaders_, 
												BOOL fUseSSL_, BSTR* pbsPage_)
{
	USES_CONVERSION;

	HINTERNET	hInet		= NULL;
	HINTERNET	hHTTP		= NULL;
	HINTERNET	hReq		= NULL;

	CONST UINT	uiChunkSize	= 10*1024;
	CHAR		cChunkBuf[uiChunkSize+1];

	LPTSTR		szVerb;

	LPSTR		szPageBuf		= NULL;
	DWORD		dwPageBufSize	= uiChunkSize*10;
	DWORD		dwDataLen		= 0;

	BOOL		fRet;
	DWORD		dwLastErr;

	cfdprintf(_T("CBotikBalanceObj::RequestPage"));

	__try
	{
		if(g_pDataCont->GetUseProxy())
		{
			cfdprintf(_T("CBotikBalanceObj::RequestPage: Check that proxy address has address:port view..."));

			if(g_pDataCont->GetProxyServer() == NULL || _tcschr(g_pDataCont->GetProxyServer(), _T(':')) == NULL)
			{
				cfdprintf(_T("CBotikBalanceObj::RequestPage: Proxy server address is incorrest... (%s)"), g_pDataCont->GetProxyServer());
				m_lLastOpCode = BBO_RET_WRONGPROXY;
				__leave;
			};
		};

		hInet = InternetOpen(_T("BOTIK-BALANCE"), g_pDataCont->GetUseProxy() ? INTERNET_OPEN_TYPE_PROXY : INTERNET_OPEN_TYPE_DIRECT, 
							 g_pDataCont->GetUseProxy() ? g_pDataCont->GetProxyServer() : NULL, NULL, 0);
		if(hInet == NULL)
		{
			cfdprintf(_T("CBotikBalanceObj::RequestPage: InternetOpen failed (%d)."), GetLastError());
			m_lLastOpCode = BBO_RET_INETOPEN;
			__leave;
		};

		hHTTP = InternetConnect(hInet, W2T(bsHost_), (WORD)lPort_, NULL, NULL, INTERNET_SERVICE_HTTP, 0, NULL);
		if(hHTTP == NULL)
		{
			cfdprintf(_T("CBotikBalanceObj::RequestPage: InternetConnect failed (%d)."), GetLastError());
			m_lLastOpCode = BBO_RET_INETCONN;
			__leave;
		};

		if(lMethod_ == BBO_HTTP_GET)
			szVerb = _T("GET");
		else if(lMethod_ == BBO_HTTP_POST)
			szVerb = _T("POST");
		else if(lMethod_ == BBO_HTTP_PUT)
			szVerb = _T("PUT");
		else
		{
			cfdprintf(_T("CBotikBalanceObj::RequestPage: Method number not supported (%d)."), lMethod_);
			m_lLastOpCode = BBO_RET_WRONGMETHOD;
			__leave;
		};

		cfdprintf(_T("CBotikBalanceObj::RequestPage: Method: %s; URL: %s"), szVerb, W2T(bsPath_));

		hReq = HttpOpenRequest(hHTTP, szVerb, W2T(bsPath_), NULL, NULL, NULL, 
							   INTERNET_FLAG_NO_UI|INTERNET_FLAG_PRAGMA_NOCACHE|INTERNET_FLAG_RELOAD|
							   INTERNET_FLAG_IGNORE_REDIRECT_TO_HTTP|INTERNET_FLAG_IGNORE_REDIRECT_TO_HTTPS|
							   INTERNET_FLAG_IGNORE_CERT_DATE_INVALID|INTERNET_FLAG_IGNORE_CERT_CN_INVALID|
							   (fUseSSL_ ? INTERNET_FLAG_SECURE : 0), NULL);

		if(hReq == NULL)
		{
			cfdprintf(_T("CBotikBalanceObj::RequestPage: HttpOpenRequest failed (%d)."), GetLastError());
			m_lLastOpCode = BBO_RET_OPENREQ;
			__leave;
		};

		LPBYTE lpBodyPtr = NULL;
		DWORD  dwBodySize = 0;
		if(bsBody_ != NULL)
		{
			lpBodyPtr  = (LPBYTE)W2T(bsBody_);
			dwBodySize = (DWORD)strlen((PCHAR)lpBodyPtr);
		};

		LPTSTR szHdrs = NULL;
		DWORD dwHdrsSize = 0;
		if(bsHeaders_ != NULL && ::SysStringLen(bsHeaders_) > 0)
		{
			szHdrs     = W2T(bsHeaders_);
			dwHdrsSize = (DWORD)_tcslen(szHdrs);
		};

		BOOL fRecall = true;
recall:
		fRet = HttpSendRequest(hReq, szHdrs, dwHdrsSize, lpBodyPtr, dwBodySize);

		dwLastErr = GetLastError();

		if(fRet != TRUE && fRecall)
		{
			if(dwLastErr == ERROR_INTERNET_INVALID_CA)
			{	// To accept invalid CA
				DWORD dwFlags;
				DWORD dwBuffLen = sizeof(dwFlags);

				cfdprintf(_T("CBotikBalanceObj::RequestPage: HttpSendRequest failed. INVALID_CA fix flags, and do one recall."));

				InternetQueryOption(hReq, INTERNET_OPTION_SECURITY_FLAGS, (LPVOID)&dwFlags, &dwBuffLen);

				dwFlags |= SECURITY_FLAG_IGNORE_UNKNOWN_CA;

				InternetSetOption(hReq, INTERNET_OPTION_SECURITY_FLAGS, &dwFlags, sizeof(dwFlags));

				fRecall = false;

				goto recall;
			};
		};

		if(fRet != TRUE)
		{
			cfdprintf(_T("CBotikBalanceObj::RequestPage: HttpSendRequest failed (%d)"), dwLastErr);
			m_lLastOpCode = BBO_RET_SENDREQ;
			__leave;
		};

		szPageBuf = new CHAR[dwPageBufSize+1];

		// Download data loop.
		while(TRUE)
		{
			DWORD dwRead;

			fRet = InternetReadFile(hReq, cChunkBuf, uiChunkSize, &dwRead);
			if(!fRet)
			{
				cfdprintf(_T("CBotikBalanceObj::RequestPage: InternetReadFile failed (%d)."), GetLastError());
				m_lLastOpCode = BBO_RET_INETREAD;
				__leave;
			};

			cfdprintf(_T("CBotikBalanceObj::RequestPage: Got %d bytes."), dwRead);

			if(dwRead == 0)
			{
				cfdprintf(_T("CBotikBalanceObj::RequestPage: Download complited."));
				break;
			}
			else
			{
				cfdprintf(_T("CBotikBalanceObj::RequestPage: Buf size: %d; Data in buf: %d; Chunk len: %d"), dwPageBufSize, dwDataLen, dwRead);

				// Check if buffer has enought space for new data.
				if(dwPageBufSize <= (dwDataLen+dwRead+1))
				{
					CHAR*	ptr			= szPageBuf;
					DWORD	dwNewSize	= dwPageBufSize*2+dwRead;

					cfdprintf(_T("CBotikBalanceObj::RequestPage: Buffer size is not enought, new size: %d"), dwNewSize);

					// Get new space for page buffer.
					szPageBuf = new CHAR[dwNewSize+1];
					memcpy(szPageBuf, ptr, dwPageBufSize);
					dwPageBufSize = dwNewSize;
					delete[] ptr;
				};

				// Copy new data into buffer.
				memcpy((szPageBuf+dwDataLen), cChunkBuf, dwRead);
				dwDataLen += dwRead;
			};
		};

		// Terminate zero...
		*(szPageBuf+dwDataLen) = 0;

		cfdprintf(_T("CBotikBalanceObj::RequestPage: Total page size: %d; Buf size: %d"), dwDataLen, dwPageBufSize);

		// Check charset.
//		if(strstr(szPageBuf, "charset=koi8-r;") != NULL) 
		if(strstr(szPageBuf, "charset=windows-1251;") == NULL) // Not all botik's pages have charset, so assume that koi-8 is default.
		{
			// Change page charset.
			cfdprintf(_T("CBotikBalanceObj::RequestPage: Change page charset"));
			koi8r_to_win((PUCHAR)szPageBuf);
		};

		// Copy downloaded page into the output buffer.
		if(dwDataLen > 0)
			*pbsPage_ = A2BSTR(szPageBuf);

		m_lLastOpCode = BBO_RET_OK;
	}

	__finally
	{
		if(hReq)  InternetCloseHandle(hReq);
		if(hHTTP) InternetCloseHandle(hHTTP);
		if(hInet) InternetCloseHandle(hInet);
		if(szPageBuf) delete[] szPageBuf;
	};

	cfdprintf(_T("!CBotikBalanceObj::RequestPage"));

	return(S_OK);
};
#undef __finally
#undef __try
#undef __leave

HRESULT __stdcall CBotikBalanceObj::DbgPrint(BSTR bsStr_)
{
	cfdprintf(_T("** SCRIPTDBG: %s"), CW2T(bsStr_));
	return(S_OK);
};

HRESULT __stdcall CBotikBalanceObj::SetData(BSTR bsName_, BSTR bsVal_)
{
	g_pDataCont->SetData(CW2T(bsName_), CW2T(bsVal_));
	return(S_OK);
};

HRESULT __stdcall CBotikBalanceObj::get_AccountName(BSTR* pbsRet_)
{
	*pbsRet_ = T2BSTR(g_pDataCont->GetAccount());
	return(S_OK);
};

HRESULT __stdcall CBotikBalanceObj::get_User(BSTR* pbsRet_)
{
	*pbsRet_ = T2BSTR(g_pDataCont->GetUser());
	return(S_OK);
};

HRESULT __stdcall CBotikBalanceObj::get_Password(BSTR* pbsRet_)
{
	*pbsRet_ = T2BSTR(g_pDataCont->GetPassword());
	return(S_OK);
};

HRESULT __stdcall CBotikBalanceObj::get_LastOperationCode(LONG* plRet_)
{
	*plRet_ = m_lLastOpCode;
	return(S_OK);
};

//////////////////////////////////////////////////////////////////////
// CBotikBalanceObj
//////////////////////////////////////////////////////////////////////

CBotikBalanceObj* CBotikBalanceObj::CreateObject()
{
	CBotikBalanceObj* pObj = NULL;
	pObj = new CBotikBalanceObj();
	return(pObj);
};

LONG CBotikBalanceObj::Fire_OnUpdate(VOID)
{
	DISPPARAMS dp;
	VARIANT    varRes;
	LONG       lRet = BBO_RET_FAIL;

	dp.rgvarg            = NULL;
	dp.cArgs             = 0;
	dp.rgdispidNamedArgs = NULL;
	dp.cNamedArgs        = 0;

	if(m_pOnUpdateSink)
	{
		m_pOnUpdateSink->Invoke(0, // Special Trident trick
								IID_NULL,
								LOCALE_SYSTEM_DEFAULT,
								DISPATCH_METHOD, &dp,
								&varRes, NULL, NULL);

		if(varRes.vt == VT_I4)
			lRet = varRes.lVal;
	};

	return(lRet);
};

LONG CBotikBalanceObj::Fire_OnConnect(VOID)
{
	DISPPARAMS dp;
	VARIANT    varRes;
	LONG       lRet = BBO_RET_FAIL;

	dp.rgvarg            = NULL;
	dp.cArgs             = 0;
	dp.rgdispidNamedArgs = NULL;
	dp.cNamedArgs        = 0;

	if(m_pOnConnectSink)
	{
		m_pOnConnectSink->Invoke(0, // Special Trident trick
								IID_NULL,
								LOCALE_SYSTEM_DEFAULT,
								DISPATCH_METHOD, &dp,
								&varRes, NULL, NULL);

		if(varRes.vt == VT_I4)
			lRet = varRes.lVal;
	};

	return(lRet);
};

LONG CBotikBalanceObj::Fire_OnDisconnect(VOID)
{
	DISPPARAMS dp;
	VARIANT    varRes;
	LONG       lRet = BBO_RET_FAIL;

	dp.rgvarg            = NULL;
	dp.cArgs             = 0;
	dp.rgdispidNamedArgs = NULL;
	dp.cNamedArgs        = 0;

	if(m_pOnDisconnectSink)
	{
		m_pOnDisconnectSink->Invoke(0, // Special Trident trick
									IID_NULL,
									LOCALE_SYSTEM_DEFAULT,
									DISPATCH_METHOD, &dp,
									&varRes, NULL, NULL);

		if(varRes.vt == VT_I4)
			lRet = varRes.lVal;
	};

	return(lRet);
};

BOOL CBotikBalanceObj::Init(ITypeLib* pITypeLib_)
{
	BOOL	fOk = FALSE;

	cfdprintf(_T("CBotikBalanceObj::Init"));

	__try
	{
		HRESULT	hr;

		// Get type information for interface object.
		hr = pITypeLib_->GetTypeInfoOfGuid(IID_IBotikBalanceObj, &m_pITypeInfo);
		pITypeLib_->Release();
		if(FAILED(hr)) 
		{
			cfdprintf(_T("CBotikBalanceObj::Init: Get type information for interface object failed (0x%x)"), hr);
			__leave;
		};

		fOk = TRUE;

		cfdprintf(_T("CBotikBalanceObj::Init: Success"));
	}

	__finally
	{
		cfdprintf(_T("CBotikBalanceObj::Init: __finaly"));
	};

	cfdprintf(_T("!CBotikBalanceObj::Init"));

	return(fOk);
};




