/*  BotikBalance - User's account balance informer for Botik Network 
 *                 for Windows 9x/ME/NT/2000/XP/2003
 *  Copyright (C) 2005 Alexander V. Melnikoff
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public Licensse as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Sorry for the long lines in the sources, I have wide screen... ;-)
 */

#include <stdafx.h>
#include <botikbalancecom.h>
#include <classfactory.h>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
CClassFactory::CClassFactory(CLSID clsid_) 
{
	cfdprintf(_T("CClassFactory::CClassFactory: DllRefCnt: %d"), g_lDllRefCnt);

	m_clsidObject	= clsid_;
	m_cRef			= 1;
	InterlockedIncrement(&g_lDllRefCnt);

	cfdprintf(_T("!CClassFactory::CClassFactory: DllRefCnt: %d"), g_lDllRefCnt);
};

CClassFactory::~CClassFactory() 
{
	cfdprintf(_T("CClassFactory::~CClassFactory: DllRefCnt: %d"), g_lDllRefCnt);

	InterlockedDecrement(&g_lDllRefCnt);

	cfdprintf(_T("!CClassFactory::~CClassFactory: DllRefCnt: %d"), g_lDllRefCnt);
};


//////////////////////////////////////////////////////////////////////
// IUnknown methods.
//////////////////////////////////////////////////////////////////////
HRESULT __stdcall CClassFactory::QueryInterface(REFIID rIID_, LPVOID *ppv_) 
{
	if (rIID_ == IID_IUnknown || rIID_ == IID_IClassFactory) 
	{
		*ppv_ = static_cast<IClassFactory*>(this);
	} 
	else 
	{
		*ppv_ = NULL;
		return E_NOINTERFACE;
	};

	reinterpret_cast<IUnknown*>(*ppv_)->AddRef();

	return(S_OK);
};


ULONG __stdcall CClassFactory::AddRef() 
{
	return(InterlockedIncrement(&m_cRef));
};


ULONG __stdcall CClassFactory::Release() 
{
	if (InterlockedDecrement(&m_cRef) == 0) 
	{
		delete this;
		return(0);
	};

	return(m_cRef);
};

//////////////////////////////////////////////////////////////////////
// IClassFactory methods.
//////////////////////////////////////////////////////////////////////
HRESULT __stdcall CClassFactory::LockServer(BOOL) 
{
	return(E_NOTIMPL);
};

HRESULT __stdcall CClassFactory::CreateInstance(LPUNKNOWN pUnknown_, REFIID rIID_, LPVOID *ppObj_) 
{
	HRESULT hr = E_FAIL;
	LPVOID	pTemp = NULL;

	*ppObj_ = NULL;
	if(pUnknown_ != NULL) 
		return(CLASS_E_NOAGGREGATION);

	// Create the proper object.
	if (m_clsidObject == CLSID_BotikBalance) 
	{
		cfdprintf(_T("CClassFactory::CreateInstance: Create CLSID_BotikBalance"));
		pTemp = new CBotikBalanceCom();

		// Initialize object and run JScript engine...
		if(!((CBotikBalanceCom*)pTemp)->Init()) // XXX
		{
			delete pTemp;
			pTemp = NULL;
		};

		if(!((CBotikBalanceCom*)pTemp)->InitScriptEngine()) // XXX
		{
			delete pTemp;
			pTemp = NULL;
		};

		if(!((CBotikBalanceCom*)pTemp)->RunBotikBalanceObjScript()) // XXX
		{
			delete pTemp;
			pTemp = NULL;
		};
	};

	if (pTemp) 
	{
		cfdprintf(_T("CClassFactory::CreateInstance: Call QueryInterface of created object"));
		hr = ((IUnknown*)pTemp)->QueryInterface(rIID_, ppObj_);
	} 
	else 
	{
		cfdprintf(_T("CClassFactory::CreateInstance: Create object failed"));
		hr = E_OUTOFMEMORY;
	};

	return(hr);
};


